# -*- coding: utf-8 -*-
from .configuration import default_configuration
from .filesystem import safe_filename
from .io import extract_directory
from .logger import Logger
from .report import HTMLReport
from .utils import get_function_meta
from collections.abc import Iterable
from time import time
import itertools
import os
import pickle
import shutil
import sys


def apply_transformation(function, counter, logger, data):
    if isinstance(function, Iterable):
        for obj in function:
            data = apply_transformation(obj, counter, logger, data)
        return data
    else:
        metadata = get_function_meta(function)
        index = next(counter)
        logger.set_index(index)
        logger.start_function(metadata, data)
        print("Applying transformation {}".format(metadata['name']))
        data = function(data, logger)
        dump_fp = os.path.join(
            logger.directory,
            "{}.".format(index) + safe_filename(metadata['name']) + ".pickle"
        )
        with open(dump_fp, "wb") as f:
            pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
        logger.end_function(metadata, data)
        return data


def system_model(data_path, config=None, show=False):
    """A system model is a set of assumptions and modeling choices that define how to take a list of unlinked and unallocated datasets, and transform these datasets into a new list of datasets which are linked and each have a single reference product.

    The system model itself is a list of functions. The definition of this list - which functions are included, and in which order - is defined by the input parameter ``config``, which can be a list of functions or a :ref:`configuration` object. The ``system_model`` does the following:

    * Extract data from the input data sources
    * Initialize a :ref:`logger` object
    * Then, for each transformation function in the provided configuration:
        * Log the transformation function start
        * Apply the transformation function
        * Save the intermediate data state
        * Log the transformation function end
    * Finally, write a report.

    """
    config = config or default_configuration
    data = extract_directory(data_path)
    logger = Logger(data)
    print("Starting Ocelot model run")
    try:
        counter = itertools.count()

        for obj in config:
            data = apply_transformation(obj, counter, logger, data)

        logger.finish()
        html = HTMLReport(logger.filepath, show)

        return logger, data
    except KeyboardInterrupt:
        print("Terminating Ocelot model run")
        print("Deleting output directory:\n{}".format(logger.directory))
        shutil.rmtree(logger.directory)
        sys.exit(1)



# -*- coding: utf-8 -*-
from .transformations import (
    dummy_transformation,
    parameterization_validity_checks,
    relabel_global_to_row,
)


class Configuration(object):
    """This is a dummy class, to be filled in with code that can parse various ways for defining a system model in a list of Python functions, including currying, etc."""
    def __init__(self):
        self.functions = []

    def __iter__(self):
        return iter(self.functions)


# Default config for now is 3.2 cutoff
default_configuration = [
    parameterization_validity_checks,
    relabel_global_to_row,
    dummy_transformation,
]



# -*- coding: utf-8 -*-
from .filesystem import get_base_output_directory, create_dir, check_dir
from time import time
import json
import os
import uuid


class Logger(object):
    """The ``Logger`` class provides a JSON logger for use during a model run, and formats a nice report afterwards.

    ``Logger`` provides methods for several types of log messages.

    """
    def __init__(self, data):
        """Initialize the log with the raw extracted data"""
        report_id = uuid.uuid4().hex
        self.directory = self.create_output_directory(report_id)
        self.filepath = os.path.join(self.directory, "report.log.json")
        print("Opening log file at: {}".format(self.filepath))
        self.logfile = open(self.filepath, "w", encoding='utf-8')
        self.log({
            'count': len(data),
            'time': time(),
            'type': 'report start',
            'uuid': report_id,
        })
        self.index = 1

    def create_output_directory(self, report_id):
        directory = os.path.join(get_base_output_directory(), report_id)
        try:
            create_dir(directory)
            assert check_dir(directory)
        except:
            raise OutputDirectoryError(
                "Can't find or write to output directory:\n\t{}".format(
                directory)
            )
        return directory

    def set_index(self, index):
        self.index = index + 1

    def log(self, message):
        self.logfile.write(json.dumps(message) + "\n")

    def start_function(self, metadata, data):
        log_data = {
            'type': 'function start',
            'count': len(data),
            'time': time(),
            'index': self.index,
        }
        log_data.update(metadata)
        self.log(log_data)

    def end_function(self, metadata, data):
        log_data = {
            'type': 'function end',
            'count': len(data),
            'time': time(),
            'index': self.index,
        }
        log_data.update(metadata)
        self.log(log_data)

    def finish(self, show=False):
        self.log({
            'type': 'report end',
            'time': time()
        })
        self.logfile.close()
        print("Generated report at: {}".format(self.filepath))



# -*- coding: utf-8 -*-
from ..collection import Collection
from ..errors import ParameterizationError


def every_exchange_with_formula_has_a_variable_name(data, logger):
    """Data validity check."""
    for ds in data:
        for exc in ds['exchanges']:
            if 'formula' in exc and 'variable' not in exc:
                raise ParameterizationError
    return data


def parameter_names_are_unique(data, logger):
    """Data validity check."""
    for ds in data:
        if not ds.get('parameters'):
            continue
        if len(ds['parameters']) != len({p['variable'] for p in ds['parameters']}):
            raise ParameterizationError
    return data


parameterization_validity_checks = Collection(
    # every_exchange_with_formula_has_a_variable_name,
    parameter_names_are_unique
)



# -*- coding: utf-8 -*-
import hashlib
import os
import re
import unicodedata
import appdirs

re_slugify = re.compile('[^\w\s-]', re.UNICODE)


def safe_filename(string):
    """Convert arbitrary strings to make them safe for filenames. Substitutes strange characters, and uses unicode normalization.

    Appends hash of `string` to avoid name collisions.

    From http://stackoverflow.com/questions/295135/turn-a-string-into-a-valid-filename-in-python"""
    safe = re.sub(
        '[-\s]+',
        '-',
        str(
            re_slugify.sub(
                '',
                unicodedata.normalize('NFKD', str(string))
            ).strip()
        )
    )
    if isinstance(string, str):
        string = string.encode("utf8")
    return safe + "." + hashlib.md5(string).hexdigest()


def create_dir(dirpath):
    """Create directory tree to `dirpath`; ignore if already exists"""
    if not os.path.isdir(dirpath):
        os.makedirs(dirpath)
    return dirpath


def check_dir(directory):
    """Returns ``True`` if given path is a directory and writeable, ``False`` otherwise."""
    return os.path.isdir(directory) and os.access(directory, os.W_OK)


def get_base_output_directory():
    """Get base directory for model run.

    Try the environment variable OCELOT_OUTPUT first, fall back to `appdirs <https://pypi.python.org/pypi/appdirs>`__"""
    try:
        return create_dir(os.environ['OCELOT_OUTPUT'])
    except KeyError:
        return create_dir(appdirs.user_data_dir("Ocelot", "ocelot_runs"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Ocelot command line interface. See https://ocelot.space/ for more info.

Usage:
  ocelot-cli run <dirpath> [--noshow]
  ocelot-cli run <dirpath> <config> [--noshow]
  ocelot-cli cleanup
  ocelot-cli validate <dirpath> <schema>
  ocelot-cli validate <dirpath>
  ocelot-cli -l | --list
  ocelot-cli -h | --help
  ocelot-cli --version

Options:
  --list        List the updates needed, but don't do anything
  --noshow      Don't open HTML report in new web browser tab
  -h --help     Show this screen.
  --version     Show version.

"""
from docopt import docopt
from ocelot import (
    cleanup_data_directory,
    data_dir,
    system_model,
    validate_directory,
)
import os
import sys


def main():
    try:
        args = docopt(__doc__, version='Ocelot open source linker CLI 0.1')
        if args['validate']:
            validate_directory(
              args['<dirpath>'],
              args['<schema>'] or os.path.join(data_dir, 'EcoSpold02.xsd')
            )
        elif args['cleanup']:
            cleanup_data_directory()
        elif args['run']:
            system_model(args["<dirpath>"], args['<config>'], show=not args['--noshow'])
        else:
            raise ValueError
    except KeyboardInterrupt:
        print("Terminating Ocelot CLI")
        sys.exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from ocelot.collection import Collection
from ocelot.model import system_model
from .mocks import fake_report

# Tests for `Collection` of functions


def test_empty_collection(fake_report):
    empty_collection = Collection()
    report, data = system_model([1,2,3,4], [empty_collection])
    assert data == [1,2,3,4]

def test_can_pass_collection_directly(fake_report):
    """i.e. not in a list"""
    collection = Collection(lambda x, y: x)
    report, data = system_model([1,2,3,4], collection)
    assert data == [1,2,3,4]

def test_can_nest_collections(fake_report):
    empty_collection = Collection()
    report, data = system_model([1,2,3,4], Collection(empty_collection))
    assert data == [1,2,3,4]

def test_empty_collection_is_falsey():
    """Passing a bare empty collection will trigger the default configuration"""
    assert not bool(Collection())

def test_collection_applied_in_order(fake_report):
    collection = Collection(
        lambda x, y: x + [1],
        lambda x, y: x + [2]
    )
    report, data = system_model([], collection)
    assert data == [1,2]

def test_list_of_functions_also_possible(fake_report):
    list_of_functions = [
        lambda x, y: x + [1],
        lambda x, y: x + [2]
    ]
    report, data = system_model([], list_of_functions)
    assert data == [1,2]

def test_collection_len():
    collection = Collection(1, 2)
    assert len(collection) == 2

def test_collection_contains():
    collection = Collection(1, 2)
    assert 1 in collection
    assert 3 not in collection



# -*- coding: utf-8 -*-
from collections.abc import Iterable


class Collection(Iterable):
    """A collection of transformation functions is correctly unwrapped by a ``system_model``.

    Useful to quickly specify a list of commonly-grouped functions (e.g. ecospold common data cleanup, economic allocation).

    Instantiate a ``Collection`` with the desired transformation functions: ``Collection(do_something, do_something_else)``.

    """
    def __init__(self, *functions):
        self.functions = functions

    def __iter__(self):
        return iter(self.functions)

    def __len__(self):
        return len(self.functions)



# -*- coding: utf-8 -*-
import os
import functools


def extract_products_as_tuple(dataset):
    return tuple(sorted([exc['name']
                  for exc in dataset['exchanges']
                  if exc['type'] == 'reference product']))


def activity_grouper(dataset):
    return (dataset['name'], extract_products_as_tuple(dataset))


def iterate_exchanges(datasets):
    for ds in datasets:
        for exc in ds['exchanges']:
            yield exc


def get_function_meta(function):
    try:
        if isinstance(function, functools.partial):
            return {
                'name': function.func.__name__,
                'description': function.func.__doc__,
                'table': getattr(function.func, "__table__", None),
            }
        else:
            return {
                'name': function.__name__,
                'description': function.__doc__,
                'table': getattr(function, "__table__", None),
            }
    except:
        raise
        return {
            'name': "callable of unknown type",
            'description': "none provided",
            'table': None,
        }



# -*- coding: utf-8 -*-
from ocelot.transformations import parameter_names_are_unique
from ocelot.errors import ParameterizationError
import pytest


def test_parameter_names_are_unique():
    data = [{'parameters': [
        {
            'name': 'foo',
            'variable': 'bar'
        }, {
            'name': 'foo',
            'variable': 'baz'
        }
    ]}]
    assert parameter_names_are_unique(data, None)


def test_parameter_names_are_unique_error():
    data = [{'parameters': [
        {
            'variable': 'foo'
        }, {
            'variable': 'foo'
        }
    ]}]
    with pytest.raises(ParameterizationError):
        parameter_names_are_unique(data, None)



# -*- coding: utf-8 -*-
SPECIAL_ACTIVITY_TYPE = {
    "0": "transforming activity",
    "1": "market activity",
    "10": "market group",
    "2": "IO activity",
    "3": "residual activity",
    "4": "production mix",
    "5": "import activity",
    "6": "supply mix",
    "7": "export activity",
    "8": "re-export activity",
    "9": "correction activity"
}

TECHNOLOGY_LEVEL = {
    "0": "undefined",
    "None": "undefined",
    "1": "new",
    "2": "modern",
    "3": "current",
    "4": "old",
    "5": "outdated"
}

INPUT_GROUPS = {
    "1": "materials/fuels",
    "2": "electricity/heat",
    "3": "services",
    "4": "from environment",
    "5": "from technosphere"
}

OUTPUT_GROUPS = {
    "0": "reference product",
    "2": "byproduct",
    "3": "material for treatment",
    "4": "to environment",
    "5": "stock addition"
}

PEDIGREE_LABELS = [
    "reliability",
    "completeness",
    "temporalCorrelation",
    "geographicalCorrelation",
    "furtherTechnologyCorrelation"
]

UNCERTAINTY_MAPPING = {
    # Translate some ecoinvent terms to their actual names
    'maxValue': 'maximum',
    'meanValue': 'mean',
    'minValue': 'minimum',
    'mostLikelyValue': 'mode',
}



# -*- coding: utf-8 -*-
from . import test_data_dir
from ocelot.io.extract_ecospold2 import generic_extractor
import os

# Tests for ecospold2 extraction


BASIC_REFERENCE = [{
    'economic': 'Business-as-Usual',
    'exchanges': [{
        'amount': -0.99,
        'id': '3b954491-a88c-484d-8332-d3941828cb4c',
        'name': 'Dr. Brayden Kilback IV',
        'production volume': {'amount': 0.0029892},
        'tag': 'intermediateExchange',
        'type': 'from technosphere',
        'uncertainty': {
            'mean': 1.,
            'mu': 2.,
            'pedigree matrix': (1, 2, 3, 4, 5),
            'type': 'lognormal',
            'variance': 3.,
            'varianceWithPedigreeUncertainty': 4.
        },
        'unit': 'm3'
    }, {
        'amount': 0.1,
        'compartments': ('air', 'unspecified'),
        'id': 'fd251456-233f-4425-aba8-38ca43bf8095',
        'name': 'Water',
        'tag': 'elementaryExchange',
        'type': 'to environment',
        'uncertainty': {
            'mean': 1.,
            'mu': 2.,
            'pedigree matrix': (1, 2, 3, 4, 5),
            'type': 'lognormal',
            'variance': 3.,
            'varianceWithPedigreeUncertainty': 4.
        },
        'unit': 'm3'
    }],
    'location': 'RoW',
    'name': 'Miss Morgan Wolf',
    'technology level': 'current',
    'temporal': ('1996-01-01', '2015-12-31'),
    'type': 'transforming activity'
}]

PROD_VOLUME_REFERENCE = [{
    'economic': 'Business-as-Usual',
    'exchanges': [{
        'amount': 1.0,
        'id': 'fb31ee9b-d01e-4121-9732-01b9bdd491ef',
        'name': 'Messiah Schiller',
        'production volume': {
            'amount': 23.0,
            'uncertainty': {
                'mean': 1.0,
                'mu': 2.0,
                'pedigree matrix': (1, 2, 3, 4, 5),
                'type': 'lognormal',
                'variance': 3.0,
                'varianceWithPedigreeUncertainty': 4.0
            },
            'variable': 'Cason_Volkman'
        },
        'tag': 'intermediateExchange',
        'type': 'reference product',
        'unit': 'kWh',
    }, {
        'amount': -1.8,
        'id': 'e1c9f077-547d-4c10-81c1-148e4a7e0df0',
        'name': 'Ellery Crooks',
        'production volume': {
            'amount': 4.0,
            'formula': 'fly_ash * apv_electricity'
        },
        'tag': 'intermediateExchange',
        'type': 'from technosphere',
        'uncertainty': {
            'mean': 1.0,
            'mu': 2.0,
            'pedigree matrix': (1, 2, 3, 4, 5),
            'type': 'lognormal',
            'variance': 3.0,
            'varianceWithPedigreeUncertainty': 4.0
        },
        'unit': 'kg',
        'variable': 'fly_ash',
    }],
    'location': 'TW',
    'name': 'Jagger Lakin',
    'technology level': 'modern',
    'temporal': ('1990-01-01', '2015-12-31'),
    'type': 'transforming activity'
}]

MULTIOUTPUT_REFERENCE = [{
    'combined production': True,
    'economic': 'Business-as-Usual',
    'exchanges': [{'amount': 200.0,
        'formula': 'carmel_reynolds + 32',
        'id': 'eae837b7-7775-4b3a-a5cd-c4c38b63dbea',
        'name': 'Dr. Jeana Considine MD',
        'production volume': {'amount': 318.0},
        'properties': [{
            'amount': 168.0,
            'id': '0ae5715b-3b9d-455e-a757-107c3ed8d6a3',
            'name': 'price',
            'variable': 'carmel_reynolds'
        }, {
            'amount': 0.6,
            'formula': 'miss_bette_kirlin',
            'id': 'a9358458-9724-4f03-b622-106eda248916',
            'name': 'water content',
            'uncertainty': {
                'mean': 1.0,
                'mu': 2.0,
                'pedigree matrix': (1, 2, 3, 4, 5),
                'type': 'lognormal',
                'variance': 3.0,
                'varianceWithPedigreeUncertainty': 4.0
            }
        }],
        'tag': 'intermediateExchange',
        'type': 'reference product',
        'unit': 'm3',
        'variable': 'pulpwood'
    }, {
        'amount': 34.0,
        'formula': 'raiden_thompson',
        'id': '667e76dc-d54f-4062-a75c-01932128f788',
        'name': 'Mr. Perley Hessel II',
        'production volume': {
            'amount': 19.0,
            'uncertainty': {
                'mean': 19.0,
                  'mu': 2.0,
                  'pedigree matrix': (1, 2, 3, 4, 5),
                  'type': 'lognormal',
                  'variance': 3.0,
                  'varianceWithPedigreeUncertainty': 4.0
            }
        },
        'tag': 'intermediateExchange',
        'type': 'reference product',
        'unit': 'kg',
        'variable': 'greenholt'}
    ],
    'location': 'GLO',
    'name': 'Aliana Price',
    'parameters': [{
        'formula': 'alia_botsford * 42',
        'name': 'Raiden Thompson IV',
        'variable': 'raiden_thompson'
    }, {
        'name': 'Mrs. Alia Botsford',
        'uncertainty': {
            'mean': 1.0,
            'mu': 2.0,
            'pedigree matrix': (1, 2, 3, 4, 5),
            'type': 'lognormal',
            'variance': 3.0,
            'varianceWithPedigreeUncertainty': 4.0
        },
        'variable': ' alia_botsford'
    }],
    'technology level': 'current',
    'temporal': ('2010-01-01', '2015-12-31'),
    'type': 'transforming activity'
}]


def test_basic_extraction():
    fp = os.path.join(test_data_dir, "basic.xml")
    BASIC_REFERENCE[0]['filepath'] = fp
    assert generic_extractor(fp) == BASIC_REFERENCE


def test_production_volume_extraction():
    fp = os.path.join(test_data_dir, "prod_volume.xml")
    PROD_VOLUME_REFERENCE[0]['filepath'] = fp
    assert generic_extractor(fp) == PROD_VOLUME_REFERENCE


def test_multioutput_extraction():
    fp = os.path.join(test_data_dir, "parameterized_multioutput.xml")
    MULTIOUTPUT_REFERENCE[0]['filepath'] = fp
    assert generic_extractor(fp) == MULTIOUTPUT_REFERENCE



# -*- coding: utf-8 -*-
from . import data_dir
import jinja2
import json
import os
import pathlib
import shutil
import webbrowser


def read_json_log(fp):
    with open(fp, encoding='utf-8') as f:
        for line in f:
            yield json.loads(line)


def jsonize(data):
    """Convert some values to JSON strings and do a few other manipulations"""
    _ = lambda x: json.dumps(x, ensure_ascii=False)
    data['count_labels'] = _([x[0] for x in data['counts']])
    data['count_data'] = _([x[1] for x in data['counts']])
    data['time_labels'] = _([x[0] for x in data['times']])
    data['time_data'] = _([x[1] for x in data['times']])
    for k, v in data['functions'].items():
        if hasattr(v, "tabledata"):
            v["tabledata"] = _(sorted(v["tabledata"]))
            v['table']['columns'] = _(v['table']['columns'])
        v['index'] = k
    data['functions'] = list(data['functions'].values())
    data['functions'].sort(key=lambda x: x['index'])
    return data


class HTMLReport(object):
    """Generate an HTML report from a :ref:`logger` logfile.

    Reports are generated in the same directory as the logfile.

    Takes the log filepath as input variable ``filepath``. A second optional input, ``show``, will open the generated report in a new webbrowser tab if ``True``.

    """
    def __init__(self, filepath, show=False):
        base_dir = os.path.abspath(os.path.dirname(filepath))
        data = self.read_log(filepath)
        self.create_assets_directory(base_dir)
        self.write_page(data, base_dir)
        self.index = 0
        if show:
            self.show_in_webbrowser()

    def show_in_webbrowser(self):
        webbrowser.open_new_tab(
            pathlib.Path(self.report_fp).as_uri()
        )

    def read_log(self, fp):
        data = {'functions': {}}
        for line in read_json_log(fp):
            self.handle_line(line, data)
        data['elapsed'] = "{:.1f}".format(data['times'][-1][1] - data['times'][0][1])
        data['times'] = [(x, y - data['times'][0][1]) for x, y in data['times']]
        return jsonize(data)

    def write_page(self, data, base_dir):
        template_filepath = os.path.join(data_dir, "report-template.jinja2")
        template = jinja2.Template(open(template_filepath).read())
        self.report_fp = os.path.join(base_dir, "report.html")
        with open(self.report_fp, "w") as f:
            f.write(template.render(**data))

    def create_assets_directory(self, base_dir):
        assets_from_dir = os.path.join(data_dir, "assets")
        assets_to_dir = os.path.join(base_dir, "assets")
        if os.path.exists(assets_to_dir):
            shutil.rmtree(assets_to_dir)
        shutil.copytree(assets_from_dir, assets_to_dir)

    def handle_line(self, line, data):
        if line['type'] == 'report start':
            data.update({
                'counts': [("Start", line['count'])],
                'times': [("Start", line['time'])],
                'uuid': line['uuid'],
            })
        elif line['type'] == 'report end':
            data['times'].append(('End', line['time']))
        elif line['type'] == 'function start':
            self.index = line['index']
            data['functions'][self.index] = {
                'start': line['time'],
            }
            if line['table']:
                data['functions'][self.index]['tabledata'] = []
        elif line['type'] == 'function end':
            data['counts'].append((line['name'], line['count']))
            data['times'].append((line['name'], line['time']))
            data['functions'][self.index].update({
                'time': line['time'] - data['functions'][self.index]['start'],
                'end': line['time'],
                'count': line['count'],
                'name': line['name'],
                'id': 'function{}'.format(self.index),
                'description': line['description'],
                'table': line['table'],
            })
        elif line['type'] == 'table element':
            data['functions'][self.index]['tabledata'].append(line['data'])




# -*- coding: utf-8 -*-
from ocelot.model import system_model
from ocelot.logger import Logger
from unittest import mock
import os
import pytest
import tempfile


class MockLogger(Logger):
    """Test mock that uses a tempfile directory, and deletes it after test runs"""
    def create_output_directory(self, report_id):
        self._tempdir = tempfile.TemporaryDirectory()
        return self._tempdir.name

    def finish(self):
        self.logfile.close()
        self._tempdir.cleanup()


def do_nothing(*args, **kwargs):
    """Mock for `HTMLReport` that doesn't do anything"""
    pass


def passthrough(obj):
    """Mock for `extract_directory` that doesn't do anything"""
    return obj


@pytest.fixture
def fake_report(monkeypatch):
    monkeypatch.setattr(
        'ocelot.model.Logger',
        MockLogger
    )
    monkeypatch.setattr(
        'ocelot.model.extract_directory',
        passthrough
    )
    monkeypatch.setattr(
        'ocelot.model.HTMLReport',
        do_nothing
    )

# Test to make sure above fixture correctly monkey-patches stuff

def test_report_and_extract_directory_mock(fake_report):
    report, data = system_model([])
    assert data == []
    # Directory not in usual place
    assert "Ocelot" not in report.directory
    # Tempdir is cleaned up
    assert not os.path.exists(report.directory)



# -*- coding: utf-8 -*-
from ..filesystem import get_base_output_directory
import os
import shutil
import time


def cleanup_data_directory():
    base_dir = get_base_output_directory()
    one_week, now, to_delete = 7 * 24 * 60 * 60, time.time(), []

    print("\nCleaning up data directory {}".format(base_dir))
    for report in os.listdir(base_dir):
        dirpath = os.path.join(base_dir, report)
        if not os.path.isdir(dirpath):
            continue
        created = os.path.getctime(dirpath)
        if created < now - one_week:
            to_delete.append((created, report, dirpath))

    if not to_delete:
        print("\nNo old reports to delete!\n")
        return
    to_delete.sort()
    template = "Deleting {count} reports:\n{items}"
    items = "".join(["\t{report} from {created}\n".format(
                     report=y, created=time.ctime(x)) for x, y, z in to_delete])
    print(template.format(count=len(to_delete), items=items))

    for _, _, dirpath in to_delete:
        shutil.rmtree(dirpath)



# -*- coding: utf-8 -*-
__all__ = (
    "Configuration",
    "HTMLReport",
    "Logger",
    "system_model",
)

__version__ = (0, 1)

CYTOOLZ = """Functional programming library cytoolz not found.

Please consider installing the cytoolz library using a package manager
or http://www.lfd.uci.edu/~gohlke/pythonlibs/ (for Windows). There
are significant speedups versus the pure-python library toolz.

See https://github.com/pytoolz/cytoolz for more information."""

try:
    import cytoolz as toolz
except ImportError:
    import warnings
    import toolz
    warnings.warn(CYTOOLZ)

from .data import data_dir
from .collection import Collection
from .io import *
from .configuration import Configuration, default_configuration
from .logger import Logger
from .report import HTMLReport
from .model import system_model



# -*- coding: utf-8 -*-
from .. import toolz
from ..utils import activity_grouper


def relabel_global_to_row(data, logger):
    """Change ``GLO`` locations to ``RoW`` if there are region-specific datasets in the activity group."""
    processed = []
    for key, datasets in toolz.groupby(activity_grouper, data).items():
        if len(datasets) > 1:
            for ds in datasets:
                if ds['location'] == 'GLO':
                    ds['location'] = 'RoW'
                    logger.log({
                        'type': 'table element',
                        'data': (key[0], "; ".join(sorted(key[1])))
                    })
                processed.append(ds)
        else:
            processed.extend(datasets)
    return processed

relabel_global_to_row.__table__ = {
    'title': 'Activities changed from `GLO` to `RoW`',
    'columns': ["Name", "Product(s)"]
}



from setuptools import setup
import os

packages = []
root_dir = os.path.dirname(__file__)
if root_dir:
    os.chdir(root_dir)

for dirpath, dirnames, filenames in os.walk('ocelot'):
    # Ignore dirnames that start with '.'
    if '__init__.py' in filenames:
        pkg = dirpath.replace(os.path.sep, '.')
        if os.path.altsep:
            pkg = pkg.replace(os.path.altsep, '.')
        packages.append(pkg)

setup(
    name='ocelot-lca',
    version="0.1",
    packages=packages,
    author="Chris Mutel",
    author_email="cmutel@gmail.com",
    license=open('LICENSE.txt').read(),
    package_data={'ocelot': [
        "data/*.*",
        "data/*/*.*",
    ]},
    entry_points = {
        'console_scripts': [
            'ocelot-cli = ocelot.bin.ocelot_cli:main',
        ]
    },
    install_requires=[
        'appdirs',
        'arrow',
        'docopt',
        'jinja2',
        'lxml',
        'psutil',
        'pyprind',
        'pyrsistent',
        'pytest',
        'toolz',
    ],
    url="https://ocelot.space/",
    long_description=open('README.md').read(),
    description='Open source linking for life cycle assessment',
    classifiers=[
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
)

# Also consider:
# http://code.activestate.com/recipes/577025-loggingwebmonitor-a-central-logging-server-and-mon/



# -*- coding: utf-8 -*-
from .locations import relabel_global_to_row
from .parameterization import (
    parameterization_validity_checks,
    parameter_names_are_unique,
)


def dummy_transformation(data, logger):
    """This is a dummy transformation that doesn't do anything.

    Used primarily for testing."""
    return data



# -*- coding: utf-8 -*-
from lxml import etree
import os
import pprint
import pyprind


def validate_directory(dirpath, schema):
    """Extract all the ``.spold`` files in the directory ``dirpath``.

    Use a multiprocessing pool if ``use_mp``, which is the default."""
    assert os.path.isdir(dirpath), "Can't find data directory {}".format(dirpath)
    assert os.path.isfile(schema), "Can't find schema file {}".format(schema)

    filelist = [os.path.join(dirpath, filename)
                for filename in os.listdir(dirpath)
                if filename.lower().endswith(".spold")
                ]

    print("Validating {} undefined datasets".format(len(filelist)))

    errors = []
    ecospold2_schema = etree.XMLSchema(etree.parse(open(schema)))

    for fp in pyprind.prog_bar(filelist):
        file = etree.parse(open(fp))
        if not ecospold2_schema.validate(file):
            errors.append(os.path.basename(fp))

    if errors:
        print("The following files did not validate:")
        pprint.pprint(errors)
    else:
        print("All files valid")



# -*- coding: utf-8 -*-
from .ecospold2_meta import *
from lxml import objectify
from time import time
import multiprocessing
import os
import pprint
import pyprind
import signal


def _(string):
    return string.replace("{http://www.EcoInvent.org/EcoSpold02}", "")


def is_combined_production(dataset):
    """"Combined production datasets have multiple reference products.

    Returns a boolean."""
    return len([1 for exc in dataset['exchanges']
                if exc['type'] == "reference product"]) > 1


def parameterize(elem, data):
    parameters = [
        obj for obj in elem.flowData.iterchildren()
        if 'parameter' in obj.tag
    ]
    if not parameters:
        return
    else:
        data['parameters'] = []
    for param in parameters:
        obj = {
            'variable': param.get('variableName'),
            'name': param.name.text,
        }
        if hasattr(param, "uncertainty"):
            obj['uncertainty'] = extract_uncertainty(param.uncertainty)
        formula = param.get('mathematicalRelation')
        if formula:
            obj['formula'] = formula
        data['parameters'].append(obj)


def extract_pedigree_matrix(elem):
    try:
        return tuple([
            int(elem.pedigreeMatrix.get(label))
            for label in PEDIGREE_LABELS
        ])
    except:
        pass


def extract_uncertainty(unc):
    distribution_labels = ['lognormal', 'normal', 'triangular',
                           'uniform', 'undefined']
    distribution = next((getattr(unc, label)
                         for label in distribution_labels
                         if hasattr(unc, label)))
    data = {UNCERTAINTY_MAPPING.get(key, key): float(distribution.get(key)) for key in distribution.keys()}
    data.update({
        'type': _(distribution.tag),
    })
    pm = extract_pedigree_matrix(unc)
    if pm:
        data['pedigree matrix'] = pm
    return data


def extract_compartments(exc):
    try:
        return (
            exc.compartment.compartment.text,
            exc.compartment.subcompartment.text,
        )
    except:
        pass


def extract_production_volume(exc):
    pv = exc.get('productionVolumeAmount')
    if not pv:
        return
    data = {'amount': float(pv)}
    if hasattr(exc, "productionVolumeUncertainty"):
        data['uncertainty'] = extract_uncertainty(exc.productionVolumeUncertainty)
    formula = exc.get('productionVolumeMathematicalRelation')
    if formula:
        data['formula'] = formula
    variable = exc.get('productionVolumeVariableName')
    if variable:
        data['variable'] = variable
    return data


def extract_property(prop):
    data = {
        'id': prop.get('propertyId'),
        'amount': float(prop.get('amount')),
        'name': prop.name.text,
    }
    if hasattr(prop, "uncertainty"):
        data['uncertainty'] = extract_uncertainty(prop.uncertainty)
    if prop.get("variableName"):
        data['variable'] = prop.get("variableName")
    if prop.get("mathematicalRelation"):
        data['formula'] = prop.get("mathematicalRelation")
    return data


def extract_minimal_exchange(exc):
    # Basic data
    data = {
        'id': exc.get('id'),
        'tag': _(exc.tag),
        'name': exc.name.text,
        'unit': exc.unitName.text,
        'amount': float(exc.get('amount')),
        'type': (INPUT_GROUPS[exc.inputGroup.text]
                 if hasattr(exc, "inputGroup")
                 else OUTPUT_GROUPS[exc.outputGroup.text]),
    }

    # Variable and formula will be deleted if dataset is not multioutput
    if exc.get("variableName"):
        data['variable'] = exc.get("variableName")
    if exc.get("mathematicalRelation"):
        data['formula'] = exc.get("mathematicalRelation")

    properties = [extract_property(obj)
                  for obj in exc.iterchildren()
                  if _(obj.tag) == 'property']
    if properties:
        data['properties'] = properties

    # Biosphere compartments
    compartments = extract_compartments(exc)
    if compartments:
        data['compartments'] = compartments

    # Production volume & uncertainty
    pv = extract_production_volume(exc)
    if pv:
        data['production volume'] = pv

    # Uncertainty
    if hasattr(exc, "uncertainty"):
        data['uncertainty'] = extract_uncertainty(exc.uncertainty)

    return data


def extract_minimal_ecospold2_info(elem, filepath):
    data = {
        'filepath': filepath,
        'name': elem.activityDescription.activity.activityName.text,
        'location': elem.activityDescription.geography.shortname.text,
        'type': SPECIAL_ACTIVITY_TYPE[elem.activityDescription.activity.get('specialActivityType')],
        'technology level': TECHNOLOGY_LEVEL[elem.activityDescription.technology.get('technologyLevel')],
        'temporal': (
            elem.activityDescription.timePeriod.get("startDate"),
            elem.activityDescription.timePeriod.get("endDate")
        ),
        'economic': elem.activityDescription.macroEconomicScenario.name.text,
        'exchanges': [extract_minimal_exchange(exc)
                      for exc in elem.flowData.iterchildren()
                      if 'Exchange' in exc.tag]
    }
    if is_combined_production(data):
         # Dataset will require recalculation, so variableName
         # and mathematicalRelation are necessary
        data['combined production'] = True
        parameterize(elem, data)
    else:
        # Don't need parameter or property data because no allocation
        for exc in data['exchanges']:
            if 'formula' in exc:
                del exc['formula']
            if 'properties' in exc:
                del exc['properties']
    return data


def generic_extractor(filepath):
    with open(filepath, encoding='utf8') as f:
        try:
            root = objectify.parse(f).getroot()
            data = [extract_minimal_ecospold2_info(child, filepath)
                    for child in root.iterchildren()]
        except:
            print(filepath)
            raise
    return data


def extract_directory(dirpath, use_mp=True):
    """Extract all the ``.spold`` files in the directory ``dirpath``.

    Use a multiprocessing pool if ``use_mp``, which is the default."""
    assert os.path.isdir(dirpath), "Can't find directory {}".format(dirpath)
    filelist = [os.path.join(dirpath, filename)
                for filename in os.listdir(dirpath)
                if filename.lower().endswith(".spold")
                ]

    print("Extracting {} undefined datasets".format(len(filelist)))

    if use_mp:
        start = time()
        # With code from
        # http://jtushman.github.io/blog/2014/01/14/python-%7C-multiprocessing-and-interrupts/
        with multiprocessing.Pool(
                processes=multiprocessing.cpu_count(),
                initializer=lambda : signal.signal(signal.SIGINT, signal.SIG_IGN)
            ) as pool:
            try:
                data = pool.map(generic_extractor, filelist)
            except KeyboardInterrupt:
                pool.terminate()
                raise KeyboardInterrupt
        print("Extracted {} undefined datasets in {:.1f} seconds".format(len(data), time() - start))
    else:
        data = [generic_extractor(fp)
                for fp in pyprind.prog_bar(filelist)]

    # Unroll lists of lists
    return [y for x in data for y in x]
